#!/usr/bin/env node

'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

require('core-js/modules/es.array.flat');
require('core-js/modules/es.array.unscopables.flat');
require('core-js/modules/es.promise');
var chalk = _interopDefault(require('chalk'));
var cosmiconfig = require('cosmiconfig');
var fs = require('fs-extra');
var fs__default = _interopDefault(fs);
var path = _interopDefault(require('path'));
var Git = _interopDefault(require('simple-git/promise'));
var fs$1 = _interopDefault(require('fs'));
var resolve = _interopDefault(require('resolve'));
var glob = _interopDefault(require('glob'));

const error = (err, code = 0) => {

  if (err instanceof Error) {
    console.error(err);
  } else {
    console.error(chalk`{red.bold ERROR:} ${err}`);
  }

  console.log("If you think this is a bug, you can report it: https://github.com/benawad/destiny/issues");
  process.exit(code);
};
const info = msg => {
  console.info(chalk`{green.bold INFO:} ${msg}`);
};
const log = msg => {
  console.log(msg);
};
const warn = msg => {
  console.warn(chalk`{yellow.bold WARN:} ${msg}`);
};
var logger = {
  error,
  info,
  log,
  warn
};

const moveFiles = async (newStructure, parentFolder) => {
  const git = Git(parentFolder);
  let isRepo = false;

  try {
    isRepo = await git.checkIsRepo();
  } catch {}

  for (const [k, newLocation] of Object.entries(newStructure)) {
    // skip globals
    if (k.includes("..")) {
      continue;
    }

    const oldAbsLocation = path.resolve(path.join(parentFolder, k));
    const newAbsLocation = path.resolve(path.join(parentFolder, newLocation));

    if (oldAbsLocation !== newAbsLocation) {
      // make folders
      fs__default.ensureDirSync(path.dirname(newAbsLocation));
      let shouldGitMv = false;

      if (isRepo) {
        // check if file is tracked in git
        try {
          await git.silent(true).raw(["ls-files", "--error-unmatch", oldAbsLocation]);
          shouldGitMv = true;
        } catch {}
      }

      if (shouldGitMv) {
        await git.mv(oldAbsLocation, newAbsLocation);
      } else {
        // move
        fs__default.renameSync(oldAbsLocation, newAbsLocation);
      }
    }
  }
};

/** Recursively removes all empty folders. */

function removeEmptyFolders(directory) {
  const files = fs$1.readdirSync(directory);
  if (!files) return fs$1.rmdirSync(directory);

  for (const filePath of files) {
    const fullPath = path.resolve(directory, filePath);
    const isDirectory = fs$1.lstatSync(fullPath).isDirectory();
    if (!isDirectory) continue;
    removeEmptyFolders(fullPath);
    const isEmpty = fs$1.readdirSync(fullPath).length === 0;
    if (isEmpty) fs$1.rmdirSync(fullPath);
  }
}

function findEdges(filePath) {
  const importRegex = /(?:(?:from|import)\s+["'](\.[^'"]*)["'])|(?:(?:require|import)\s*\(["'](\.[^'"]*)["']\))/gm;
  const commentRegex = /\/\*[^]*?\*\/|^.*\/\/.*$/gm;
  const edges = [];
  const fileContent = fs$1.readFileSync(filePath, {
    encoding: "utf8"
  }).toString().replace(commentRegex, "");
  let matches;

  while ((matches = importRegex.exec(fileContent)) !== null) {
    var _matches$;

    // This is necessary to avoid infinite loops with zero-width matches.
    if (matches.index === importRegex.lastIndex) {
      importRegex.lastIndex++;
    }

    edges.push([filePath, (_matches$ = matches[1]) != null ? _matches$ : matches[2]]);
  }

  return edges;
}

function getExtensionFromImport(relativePath) {
  const ext = path.extname(relativePath);
  const includeExtension = [".js", ".jsx", ".ts", ".tsx"].includes(ext);
  return includeExtension ? ext : undefined;
}

const makeImportPath = (fromPath, toPath, useForwardSlashes) => {
  const fromDirectory = path.dirname(fromPath);
  const relativePath = path.relative(fromDirectory, toPath);
  const relativeDirectory = path.dirname(relativePath);
  const ext = getExtensionFromImport(relativePath);
  const fileName = path.basename(relativePath, ext);
  let newImport = path.join(relativeDirectory, fileName); // Ensures relative imports.

  const notRelative = !newImport.startsWith(".");

  if (notRelative) {
    newImport = "./" + newImport;
  } // Replace / and \ with \\.


  if (!useForwardSlashes) {
    newImport = newImport.replace(/\/|\\+/g, "\\\\");
  } // Replace \ with /.
  else {
      newImport = newImport.replace(/\\/g, "/");
    }

  return newImport;
};

/** Resolve with a list of predefined extensions. */

const customResolve = (id, basedir) => {
  return resolve.sync(id, {
    basedir,
    extensions: [".js", ".jsx", ".sass", ".scss", ".svg", ".ts", ".tsx"]
  });
};

const getNewFilePath = (file, rootOptions) => {
  for (const {
    tree,
    parentFolder
  } of rootOptions) {
    const key = path.relative(parentFolder, file);

    if (key in tree) {
      return path.resolve(path.join(parentFolder, tree[key]));
    }
  }

  return file;
};

const getNewImportPath = (absImportPath, newFilePath, rootOptions) => {
  let lastUseForwardSlash = true;

  for (const {
    tree,
    parentFolder,
    useForwardSlash
  } of rootOptions) {
    lastUseForwardSlash = useForwardSlash;
    const key = path.relative(parentFolder, absImportPath);

    if (key in tree) {
      return makeImportPath(newFilePath, path.resolve(path.join(parentFolder, tree[key])), useForwardSlash);
    }
  }

  return makeImportPath(newFilePath, absImportPath, lastUseForwardSlash);
};

const fixImports = (filePaths, rootOptions) => {
  for (const filePath of filePaths) {
    const imports = findEdges(filePath);
    if (!imports.length) continue;
    const basedir = path.dirname(filePath);
    const newFilePath = getNewFilePath(filePath, rootOptions);
    const ogText = fs.readFileSync(filePath).toString();
    let newText = ogText.repeat(1);

    for (const imp of imports) {
      const absPath = customResolve(imp[1], basedir);
      const newImportText = getNewImportPath(absPath, newFilePath, rootOptions);

      if (newImportText) {
        newText = newText.replace(`'${imp[1]}'`, `'${newImportText}'`).replace(`"${imp[1]}"`, `"${newImportText}"`);
      }
    }

    if (newText !== ogText) {
      fs.writeFileSync(filePath, newText);
    }
  }
};

const formatFileStructure = async (filePaths, rootOptions) => {
  logger.info("Fixing imports.");
  fixImports(filePaths, rootOptions);

  for (const {
    tree,
    parentFolder
  } of rootOptions) {
    logger.info("Moving files.");
    await moveFiles(tree, parentFolder);
    removeEmptyFolders(parentFolder);
  }

  logger.info("Restructure was successful!");
};

const createBranchFromParts = (parts, count) => parts.slice(0, count).join("/");
/** Remove path that matches `match` but save '/' to calculate position. */


const removePathDuplication = (target, match) => target.replace(new RegExp(`^(/*)${match.replace(/\//g, "")}(/+)`), "$1$2");
/**
 * Matches anything between '/' and '.' and prepends the highest possible char
 * code to it. This enables us sort files lower than directories.
 */


const prependMaxCharCodeToFile = text => text.replace(/([^/]+)(?=\.)/g, String.fromCharCode(Number.MAX_SAFE_INTEGER) + "$1");

const compareLeafs = (a, b) => prependMaxCharCodeToFile(a).localeCompare(prependMaxCharCodeToFile(b));
/** Check if leaf is the last of its siblings excluding children. */


const isLeafLastSibling = (leaf, remainingLeafs) => {
  for (const remaningLeaf of remainingLeafs) {
    if (remaningLeaf.position > leaf.position) continue;
    if (remaningLeaf.position < leaf.position) return true;
    return false;
  }

  return true;
};

const removeIllicitIndentGuidelines = (line, pastLine) => {
  const isCharEndConnectorOrWhitespace = char => char === "└" || char === " ";

  return Array.from(line).map((char, idx) => char === "│" && isCharEndConnectorOrWhitespace(pastLine[idx]) ? " " : char).join("");
};
/** Resolve every unique leaf from a list of paths. */


const resolveLeafs = paths => {
  const leafs = paths.reduce((acc, target) => {
    const parts = target.split("/");
    parts.forEach((_, idx) => acc.add(createBranchFromParts(parts, idx + 1)));
    return acc;
  }, new Set());
  return Array.from(leafs).sort(compareLeafs);
};
/**
 * Iterates over all leafs and positions them on the correct branches, ie.
 * resolving their end name in relation to their branch and their position.
 */


const positionLeafs = leafs => {
  const res = [];
  let queue = [...leafs];

  while (queue.length > 0) {
    const leaf = queue.shift();
    if (leaf == null) break;
    queue = queue.map(queuedLeaf => removePathDuplication(queuedLeaf, leaf));
    res.push(leaf);
  }

  return res.map(x => {
    var _parts$pop;

    const parts = x.split("/");
    return {
      text: (_parts$pop = parts.pop()) != null ? _parts$pop : "",
      position: parts.length
    };
  });
};
/** Print a visualization of a tree of paths. */


const printTree = paths => {
  const leafs = resolveLeafs(paths);
  const positionedLeafs = positionLeafs(leafs);
  const treeVisualization = positionedLeafs.reduce((lines, currentLeaf, idx) => {
    var _lines;

    const pastLine = (_lines = lines[idx - 1]) != null ? _lines : "";
    const remainingLeafs = positionedLeafs.slice(idx + 1);
    const isDirectory = remainingLeafs.length > 0 && remainingLeafs[0].position > currentLeaf.position;
    const indent = currentLeaf.position > 0 ? "│  ".repeat(currentLeaf.position) : "";
    const connector = isLeafLastSibling(currentLeaf, remainingLeafs) ? "└──" : "├──";
    const text = isDirectory ? chalk.bold.blue(currentLeaf.text) : currentLeaf.text;
    const line = indent + connector + text;
    lines.push(removeIllicitIndentGuidelines(line, pastLine));
    return lines;
  }, []).join("\n");
  logger.log(treeVisualization);
  return treeVisualization;
};

/** Find the common parent directory between all paths. */

const findSharedParent = paths => {
  if (paths.length === 1) return path.dirname(paths[0]);
  const parentPaths = [];
  const parts = paths.map(x => x.split("/"));
  const firstParts = parts[0];
  firstParts.forEach((part, idx) => {
    const allPartsMatch = parts.every(p => p[idx] === part);

    if (allPartsMatch) {
      parentPaths.push(part);
    }
  });
  return parentPaths.join("/");
};

function addEdgeToGraph([start, end], graph) {
  if (!(start in graph)) {
    graph[start] = [];
  }

  graph[start].push(end);
}

function buildGraph(files) {
  const parentFolder = findSharedParent(files);
  const graph = {};
  const oldGraph = {};
  const totalFiles = [];
  let numForwardSlashes = 0;
  let numBackSlashes = 0;

  for (let file of files) {
    if (file === ".git") {
      continue;
    }

    file = path.resolve(file);
    const start = path.relative(parentFolder, file);

    if (!(start in oldGraph)) {
      oldGraph[start] = {
        oldLocation: file,
        imports: []
      };
    }

    totalFiles.push(start);
    findEdges(file).forEach(edge => {
      if (edge[1].includes("/")) {
        numForwardSlashes++;
      } else if (edge[1].includes("\\")) {
        numBackSlashes++;
      }

      const pathWithExtension = customResolve(edge[1], path.dirname(edge[0]));
      const end = path.relative(parentFolder, pathWithExtension);
      addEdgeToGraph([start, end], graph);
      oldGraph[start].imports.push({
        text: edge[1],
        resolved: end
      });
    });
  }

  return {
    parentFolder,
    graph,
    files: totalFiles,
    oldGraph,
    useForwardSlash: numForwardSlashes >= numBackSlashes
  };
}

const isTestFile = f => /\.test\.|\.spec\.|\.stories\./.test(f);

const invertGraph = graph => {
  const invertedGraph = {};
  Object.keys(graph).forEach(k => {
    graph[k].forEach(v => {
      if (!(v in invertedGraph)) {
        invertedGraph[v] = [];
      }

      invertedGraph[v].push(k);
    });
  });
  return invertedGraph;
};

function findEntryPoints(graph) {
  const invertedGraph = invertGraph(graph);
  const possibleEntryPoints = Object.keys(graph);
  const entryPoints = possibleEntryPoints.filter(file => {
    const importedBy = invertedGraph[file];

    if (!importedBy || !importedBy.length) {
      return true;
    }

    return importedBy.every(x => isTestFile(x));
  });

  if (entryPoints.length) {
    return entryPoints;
  }

  const levelMap = {};
  possibleEntryPoints.forEach(k => {
    const n = k.split("/").length;

    if (!(n in levelMap)) {
      levelMap[n] = [];
    }

    levelMap[n].push(k);
  });

  for (let i = 1; i < 10; i++) {
    if (i in levelMap) {
      return levelMap[i];
    }
  }

  return [];
}

const hasCycle = (node, graph, visited) => {
  const edges = graph[node];

  if (visited.has(node)) {
    return [...visited, node];
  }

  visited.add(node);
  if (edges == null || edges.length === 0) return null;

  for (const edge of edges) {
    const cycle = hasCycle(edge, graph, new Set(visited));
    if (cycle) return cycle;
  }

  return null;
};

function toFractalTree(graph, entryPoints) {
  const res = {};
  const dependencies = {};
  const testFiles = new Set();
  let containsCycle = false;

  const addDependency = (key, location) => {
    if (!Array.isArray(dependencies[key])) {
      dependencies[key] = [];
    }

    dependencies[key].push(location);
  };

  const checkDuplicates = (location, dirname, filePath) => {
    var _newLocation;

    let newLocation;

    if (Object.values(res).includes(location)) {
      newLocation = path.join(dirname, filePath.replace(/\//g, "-"));
      logger.info(`File renamed: ${filePath} -> ${newLocation}`);
    }

    return (_newLocation = newLocation) != null ? _newLocation : location;
  };

  const fn = (filePath, folderPath, graph) => {
    const basenameWithExt = path.basename(filePath);

    if (isTestFile(basenameWithExt)) {
      testFiles.add(filePath);
      return;
    }

    let folderName = path.basename(filePath, path.extname(filePath));
    const upperFolder = path.basename(path.dirname(filePath));
    const isGlobal = filePath.includes("..");
    let location = isGlobal ? filePath : path.join(folderPath, folderName === "index" && upperFolder && upperFolder !== "." ? upperFolder + path.extname(filePath) : basenameWithExt);
    location = checkDuplicates(location, folderPath, filePath);
    folderName = path.basename(location, path.extname(location));

    if (!isGlobal) {
      res[filePath] = location;
    }

    const imports = graph[filePath];

    if ((imports == null ? void 0 : imports.length) > 0) {
      const newDestination = path.join(folderPath, folderName);

      for (const importFilePath of imports) {
        if (importFilePath in res) {
          const cycle = hasCycle(importFilePath, graph, new Set());

          if (cycle) {
            containsCycle = true;
            logger.warn(`Dependency cycle detected: ${cycle.join(" -> ")}`);
          } else {
            addDependency(importFilePath, location);
          }

          continue;
        }

        addDependency(importFilePath, location);
        fn(importFilePath, newDestination, graph);
      }
    }
  };

  for (const filePath of entryPoints) {
    fn(filePath, "", graph);
  }

  if (!containsCycle) {
    Object.entries(dependencies).forEach(([k, v]) => {
      if (v.length > 1 && !k.includes("..")) {
        const parent = findSharedParent(v);
        const filename = path.basename(k);
        const upperFolder = path.basename(path.dirname(k));
        res[k] = path.join(parent, "shared", path.basename(filename, path.extname(filename)) === "index" && upperFolder && upperFolder !== "." ? upperFolder + path.extname(filename) : filename);
      }
    });
  }

  if (testFiles.size > 0) {

    for (const testFile of testFiles) {
      const [firstRelativeImport] = graph[testFile];

      if (!firstRelativeImport) {
        continue;
      }

      const testFilePath = res[firstRelativeImport];

      if (testFilePath) {
        let location = path.join(path.dirname(testFilePath), path.basename(testFile));
        location = checkDuplicates(location, path.dirname(testFilePath), testFile);
        res[testFile] = location;
      }
    }
  }

  return res;
}

function generateTrees(restructureMap) {
  return Object.entries(restructureMap).reduce((rootOptions, [rootPath, filePaths]) => {
    if (filePaths.length <= 1) return rootOptions;
    logger.info(`Generating tree for: ${rootPath}`);
    const {
      graph,
      files,
      useForwardSlash,
      parentFolder
    } = buildGraph(filePaths);
    const tree = toFractalTree(graph, findEntryPoints(graph));
    const usedFilePaths = new Set(Object.entries(graph).flat(2));
    const unusedFiles = files.filter(filePath => !usedFilePaths.has(filePath));
    logger.log(chalk.bold.blue(rootPath));
    printTree(Object.values(tree));

    if (unusedFiles.length > 0) {
      logger.warn(`Found ${unusedFiles.length} unused files:` + "\n" + unusedFiles.join("\n"));
    }

    rootOptions.push({
      parentFolder,
      tree,
      useForwardSlash
    });
    return rootOptions;
  }, []);
}

var version = "0.3.0";

const isDirectory = filePath => fs.lstatSync(filePath).isDirectory();

const isFile = filePath => fs.lstatSync(filePath).isFile();

const globSearch = pattern => {
  const matches = glob.sync(pattern);
  const files = matches.filter(match => isFile(match));

  if (files.length === 0) {
    logger.error("Could not find any files for: " + pattern, 1);
  }

  return files;
};
/** Recursively get all file paths. */


const getFilePaths = rootPath => {
  const filePaths = [];
  const paths = [rootPath];

  while (paths.length > 0) {
    const filePath = paths.shift();
    if (filePath == null || filePath.length === 0) continue;
    const isGlobPattern = glob.hasMagic(filePath);

    if (isGlobPattern) {
      filePaths.push(...globSearch(filePath));
      continue;
    }

    if (fs.existsSync(filePath)) {
      if (isFile(filePath)) {
        filePaths.push(filePath);
      } else if (isDirectory(filePath)) {
        paths.push(path.join(filePath, "/**/*.*"));
      }
    } else {
      logger.error(`Unable to resolve the path: ${filePath}`);
    }
  }

  return filePaths;
};
/** Get a restructure map with rootPath keys and filePaths values. */


const getRestructureMap = rootPaths => rootPaths.reduce((acc, rootPath) => ({ ...acc,
  [rootPath]: getFilePaths(rootPath)
}), {});

const {
  argv
} = process;
const defaultOptions = {
  help: false,
  version: false,
  write: false
};

const printVersion = () => console.log("v" + version);

const printHelp = exitCode => {
  console.log(chalk`{blue destiny} - Prettier for file structures.

{bold USAGE}

  {blue destiny} [option...] [{underline path}]

  The {underline path} argument can consist of either a {bold file path} or a {bold glob}.

{bold OPTIONS}

  -V, --version            output version number
  -h, --help               output usage information
  -w, --write              restructure and edit folders and files
  `);
  return process.exit(exitCode);
};

const parseArgs = args => args.reduce((acc, arg) => {
  switch (arg) {
    case "-h":
    case "--help":
      acc.options.help = true;
      break;

    case "-V":
    case "--version":
      acc.options.version = true;
      break;

    case "-w":
    case "--write":
      acc.options.write = true;
      break;

    default:
      acc.rootPaths.push(arg);
  }

  return acc;
}, {
  options: {},
  rootPaths: []
});

const run = async args => {
  var _ref, _cosmiconfigSync$sear;

  const config = (_ref = (_cosmiconfigSync$sear = cosmiconfig.cosmiconfigSync("destiny").search()) == null ? void 0 : _cosmiconfigSync$sear.config) != null ? _ref : {};
  const {
    options,
    rootPaths
  } = parseArgs(args);
  const mergedOptions = { ...defaultOptions,
    ...config,
    ...options
  };
  if (mergedOptions.help) return printHelp(0);
  if (mergedOptions.version) return printVersion();
  if (rootPaths.length === 0) return printHelp(1);
  logger.info("Resolving files.");
  const restructureMap = getRestructureMap(rootPaths);
  const filesToEdit = Object.values(restructureMap).flat();

  if (filesToEdit.length === 0) {
    logger.error("Could not find any files to restructure", 1);
    return;
  }

  const rootOptions = generateTrees(restructureMap);

  if (mergedOptions.write) {
    await formatFileStructure(filesToEdit, rootOptions);
  }
};

{
  run(argv.slice(2, argv.length));
}

exports.run = run;
